/* edcrypt.
 *
 *       .  -  
 *  ,   ,   
 *   .    .  Char.
 */


#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>     /* everything... */
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/poll.h>

#include "edcrypt.h"

#ifndef min
#  define min(a,b) ((a)<(b) ? (a) : (b))
#endif


typedef struct ED_driver_data
{
    uid_t uid_owner;                /* uid    */
    wait_queue_head_t inq, outq;    /*     */
    char *buffer, *end;             /* /  */
    int buffersize;                 /*      */
    char *rp, *wp;                  /*  /   */
    int nreaders, nwriters;         /*    / */
    struct semaphore sem;           /* - (mutex-exclusive)  */
} ED_driver_data;

ED_driver_data *eddata;

typedef struct ED_device_data
{
    char type;                      /* encrypt or decrypt */
    char openned;		    /* open or not oppened */
} ED_device_data;

ED_device_data eddev[EDCRYPT_MAX_DEVICES];

static struct encrypt_key_info encrypt_key;


/*******         *******/
int major = 60;
int debug = 2;
int edv_buffer = 4096;   /*   */

MODULE_PARM(major, "i");
MODULE_PARM(debug, "i");
MODULE_PARM(edv_buffer,"i");



/*******      XOR *******/
static
void transfer(int cmd, char *en_buf, char * de_buf, int size)
{
    char	*in, *out, *key;
    int	i, keysize;

    if(debug)
        printk("transfer(%d) size %d\n", cmd, size);

    if (encrypt_key.size)
    {
        if (cmd == DEV_DECRYPT)
        {
            in = en_buf;
            out = de_buf;
        }
        else
        {
            in = de_buf;
            out = en_buf;
        }

        key = encrypt_key.key;
        keysize = encrypt_key.size;

        for (i = 0; i < size; i++)
        {
            *out++ = *in++ ^ key[(i & 0x1ff) % keysize];
        }
    }
}


static
int ed_open(struct inode *inode, struct file *filp)
{
    ED_device_data * dev;
    int minor = MINOR(inode->i_rdev);

    if(debug)
        printk("ed_open(%d)\n", minor);

    if (!filp->private_data)
    {
        if (minor >= EDCRYPT_MAX_DEVICES)
            return -ENODEV;
        dev = &eddev[minor];
        filp->private_data = dev;
    }
    else
    {
        dev = filp->private_data;
    }

    if (down_interruptible(&eddata->sem))
        return -ERESTARTSYS;

    if (dev->openned)
    {
        up(&eddata->sem);
        return -EBUSY; /*   */
    }

    if (eddev[minor^1].openned && /*    ,  uid   */
            (eddata->uid_owner != current->uid) &&  /*   */
            (eddata->uid_owner != current->euid) && /*     su */
            !capable(CAP_DAC_OVERRIDE))	/*  root */
    {
        up(&eddata->sem);
        return -EBUSY;
    }

    if (!dev->openned && !eddev[minor^1].openned)
        eddata->uid_owner = current->uid; /*   open,  user_id   */

    dev->openned++;

    if (!eddata->buffer)
    { /*   */
        eddata->buffer = kmalloc(edv_buffer, GFP_KERNEL);
        if (!eddata->buffer)
        {
            up(&eddata->sem);
            return -ENOMEM;
        }
    }
    eddata->buffersize = edv_buffer;
    eddata->end = eddata->buffer + eddata->buffersize;
    eddata->rp = eddata->wp = eddata->buffer; /* rd and wr from the beginning */

    /* use f_mode,not  f_flags: it's cleaner (fs/open.c tells why) */
    if (filp->f_mode & FMODE_READ)
        eddata->nreaders++;
    if (filp->f_mode & FMODE_WRITE)
        eddata->nwriters++;
    up(&eddata->sem);

    filp->private_data = dev;

    /*   ,     .*/
    MOD_INC_USE_COUNT;
    return 0;
}


static
int ed_release(struct inode *inode, struct file *filp)
{
    ED_device_data *dev = filp->private_data;

    down(&eddata->sem);
    if (filp->f_mode & FMODE_READ)
        eddata->nreaders--;
    if (filp->f_mode & FMODE_WRITE)
        eddata->nwriters--;

    if ((eddata->nreaders + eddata->nwriters) == 0)
    {
        kfree(eddata->buffer); /*  ,     */
        eddata->buffer = NULL; /*        */
    }

    dev->openned--; /*      */
    up(&eddata->sem);

    /*    .
     *   (rmmod edcrypt)  ,  MOD_IN_USE == 0
     * (.  Used    lsmod   edcrypt)
     */
    MOD_DEC_USE_COUNT;
    return 0;
}

static
ssize_t ed_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
    ED_device_data *dev = filp->private_data;

    /*        (man pread) */
    if (f_pos != &filp->f_pos)
        return -ESPIPE; /* Illegal seek - 'is a pipe' */

    if (down_interruptible(&eddata->sem))
        return -ERESTARTSYS;

    while (eddata->rp == eddata->wp)
    { /*     */
        up(&eddata->sem); /*   */

        if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;

        if (debug)
            printk("\"%s\" reading: going to sleep\n", current->comm);

        /*   ,  ,       */
        if (wait_event_interruptible(eddata->inq, (eddata->rp != eddata->wp)))
            return -ERESTARTSYS; /* signal: tell the fs layer to handle it */

        if (down_interruptible(&eddata->sem))
            return -ERESTARTSYS;
    }
    /* ok,   */
    if (eddata->wp > eddata->rp)
        count = min(count, eddata->wp - eddata->rp);
    else
        count = min(count, eddata->end - eddata->rp);

    /*    */
    transfer(dev->type, eddata->rp, eddata->rp, count);

    if (copy_to_user(buf, eddata->rp, count))
    {
        up (&eddata->sem);
        return -EFAULT;
    }
    eddata->rp += count;
    if (eddata->rp == eddata->end)
        eddata->rp = eddata->buffer; /*  */
    up (&eddata->sem);

    /* ,    */
    wake_up_interruptible(&eddata->outq);

    if(debug)
        printk("\"%s\" did read %li bytes\n",current->comm, (long)count);

    return count;
}

static
inline int spacefree( void )
{
    if (eddata->rp == eddata->wp)
        return eddata->buffersize - 1;
    return ((eddata->rp + eddata->buffersize - eddata->wp) % eddata->buffersize) - 1;
}


static
ssize_t ed_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
    /*        (man pwrite) */
    if (f_pos != &filp->f_pos)
        return -ESPIPE; /* Illegal seek - 'is a pipe' */

    /*   :
     *     (sem==0)-   ,
     * ,   (sem=0)     .
     *  down_interruptible     ,  
     * ,   -ERESTARTSYS    .
     */
    if (down_interruptible(&eddata->sem))
        return -ERESTARTSYS;

    /*       */
    while (spacefree() == 0)
    { /*  ,    */
        up(&eddata->sem);

        /*     -     EAGAIN */
        if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;

        if (debug)
            printk("\"%s\" writing: going to sleep\n",current->comm);

        /*   ,  ,     */
        if (wait_event_interruptible(eddata->outq, spacefree() > 0))
            return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
        if (down_interruptible(&eddata->sem))
            return -ERESTARTSYS;
    }
    /* ok,     ,     */
    count = min(count, spacefree());
    if (eddata->wp >= eddata->rp)
        count = min(count, eddata->end - eddata->wp);
    else
        count = min(count, eddata->rp - eddata->wp - 1);

    if (debug)
        printk("Going to accept %li bytes to %p from %p\n",
               (long)count, eddata->wp, buf);

    /*      */
    if (copy_from_user(eddata->wp, buf, count))
    {
        up (&eddata->sem);
        return -EFAULT;
    }

    eddata->wp += count;
    if (eddata->wp == eddata->end)
        eddata->wp = eddata->buffer;

    /*   */
    up(&eddata->sem);

    /* ,   ,   read()  select() */
    wake_up_interruptible(&eddata->inq);

    if (debug)
        printk("\"%s\" did write %li bytes\n",current->comm, (long)count);

    return count;
}

static
int ed_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int minor = MINOR(inode->i_rdev);
    struct encrypt_key_info new_key;
    int ret = 0;

    if(minor>=0 && minor<=1) /*      0  1 */
    {
        switch (cmd)
        {
        case EDCRYPT_SET_KEY: /*    */
            if (copy_from_user(&new_key, (void *)arg, sizeof(new_key)))
                return -EFAULT;

            if ((unsigned int) new_key.size > sizeof(new_key.key))
                return -EINVAL;

            memset(&encrypt_key, 0, sizeof(encrypt_key)); /*    */
            memcpy(&encrypt_key, &new_key, sizeof(encrypt_key));

            if(debug)
                printk("ed_ioctl(%d), newkey %s, size %d\n", minor, new_key.key, new_key.size);
            break;
        default:
            ret = -ENOTTY; /*   ioctl()  */
        }
    } else ret = -ENOTTY;

    return ret;
}

static unsigned int ed_poll(struct file *filp, poll_table *wait)
{
    unsigned int mask = 0;

    /*  ;    */
    int left = (eddata->rp + eddata->buffersize - eddata->wp) % eddata->buffersize;

    poll_wait(filp, &eddata->inq,  wait);
    poll_wait(filp, &eddata->outq, wait);
    /*   ,    select() */
    if (eddata->rp != eddata->wp) mask |= POLLIN | POLLRDNORM;  /*  */
    if (left != 1)          mask |= POLLOUT | POLLWRNORM; /*  */

    return mask;
}

static
void ed_proc_offset(char *buf, char **start, off_t *offset, int *len)
{
    if (*offset == 0)
        return;
    if (*offset >= *len)
    {
        *offset -= *len;
        *len = 0;
    }
    else
    {
        *start = buf + *offset;
        *offset = 0;
    }
}

static
int ed_read_status(char *buf, char **start, off_t offset, int count, int unused)
{
#define LIMIT (PAGE_SIZE-200) /*      */
    int i, len;

    *start = buf;
    len = sprintf(buf,"edcrypt "VERSION" info:\n");
    len += sprintf(buf+len, "Encryption: %s\n", (encrypt_key.size)? "Yes":"No. Please set password by edkey programm.");
    if ((eddata->nreaders + eddata->nwriters) != 0)
        len += sprintf(buf+len, "Encryption UserID ownered: %d\n", eddata->uid_owner);
    len += sprintf(buf+len, "key :[%s], size %d\n", encrypt_key.key, encrypt_key.size);
    len += sprintf(buf+len, "buffersize:%i\n\n", edv_buffer);

    if (down_interruptible(&eddata->sem))
        return -ERESTARTSYS;

    for (i = 0; i < EDCRYPT_MAX_DEVICES; i++)
    {
        len += sprintf(buf+len, "Device %i (/dev/%s): %s\n",
                       i, (eddev[i].type == DEV_ENCRYPT)? "edcrypt":"decrypt",
                       (eddev[0].openned) ? "openned":"closed");
    }

    len += sprintf(buf+len, "\nQueues: %p %p\n", eddata->inq, eddata->outq);
    len += sprintf(buf+len, "Buffer: %p to %p (%i bytes)\n",
                   eddata->buffer, eddata->end, eddata->buffersize);
    len += sprintf(buf+len, "rp %p   wp %p\n", eddata->rp, eddata->wp);
    len += sprintf(buf+len, "readers %i   writers %i\n",
                   eddata->nreaders, eddata->nwriters);

    up(&eddata->sem);
    ed_proc_offset(buf, start, &offset, &len);

    return len;
}

static struct proc_dir_entry ed_proc_status =
    {
        0,       		/* low_ino: the inode -- dynamic */
        7, "edcrypt",		/* len of name and name */
        S_IFREG | S_IRUGO,	/* mode */
        1, 0, 0,		/* nlinks, owner, group */
        0, NULL,		/* size - unused; operations -- use default */
        &ed_read_status,	/* function used to read data */
    };

static struct file_operations ed_fops =
    {
    read:	ed_read,
    write:	ed_write,
    poll:	ed_poll,
    ioctl:	ed_ioctl,
    open:	ed_open,
    release:	ed_release
    };


int init_module(void)
{
    /*      */
    if(register_chrdev(major,"edcrypt",&ed_fops) == -EBUSY)
    {
        printk("Unable to get major %d for edcrypt module!\n",major);
        return -EIO;
    }

    /*    */
    eddata = kmalloc(EDCRYPT_MAX_DEVICES * sizeof(struct encrypt_key_info), GFP_KERNEL);
    if (eddata == NULL)
        return -ENOMEM;

    memset(eddata, 0, EDCRYPT_MAX_DEVICES * sizeof(struct encrypt_key_info));

    eddev[0].type = DEV_ENCRYPT;
    eddev[1].type = DEV_DECRYPT;

    memset(&encrypt_key, 0, sizeof(encrypt_key));

    /*  read/write     */
    init_waitqueue_head(&eddata->inq);
    init_waitqueue_head(&eddata->outq);
    sema_init(&eddata->sem, 1);

    /*    /proc/edcrypt */
    proc_register(&proc_root, &ed_proc_status);

    printk("edcrypt "VERSION" module loaded.\n");

    return 0;
}

void cleanup_module(void)
{
    /*       */
    unregister_chrdev(major, "edcrypt");

    /*       /proc/edcrypt */
    proc_unregister(&proc_root, ed_proc_status.low_ino);

    if (eddata)
        kfree(eddata);

    if(debug)
        printk("edcrypt module cleanup.\n");
}

